@@ -50,10 +50,11 @@ use Coro;
 use Coro::AnyEvent qw(sleep);
 use Coro::Channel;
 use Coro::ProcessPool::Process;
-use Coro::ProcessPool::Util qw(cpu_count);
+use Coro::ProcessPool::Util;
 use Coro::Semaphore;
+require Coro::ProcessPool::Pipeline;
 
-our $VERSION = '0.18_5';
+our $VERSION = '0.22_2';
 
 if ($^O eq 'MSWin32') {
     die 'MSWin32 is not supported';
@@ -71,7 +72,7 @@ to the number of CPUs on the ssytem.
 has max_procs => (
     is      => 'ro',
     isa     => Int,
-    default => \&cpu_count,
+    default => sub { Coro::ProcessPool::Util::cpu_count() },
 );
 
 =head2 max_reqs
@@ -85,7 +86,7 @@ memory over time.
 has max_reqs => (
     is      => 'ro',
     isa     => Int,
-    default => 0,
+    default => sub { 0 },
 );
 
 =head1 PRIVATE ATTRIBUTES
@@ -98,8 +99,8 @@ to the number of processes (C<max_procs>).
 =cut
 
 has procs_lock => (
-    is      => 'lazy',
-    isa     => InstanceOf['Coro::Semaphore'],
+    is  => 'lazy',
+    isa => InstanceOf['Coro::Semaphore'],
 );
 
 sub _build_procs_lock {
@@ -116,7 +117,7 @@ Running total of processes that are currently running.
 has num_procs => (
     is      => 'rw',
     isa     => Int,
-    default => 0,
+    default => sub { 0 },
 );
 
 =head2 procs
@@ -131,61 +132,6 @@ has procs => (
     default => sub { [] },
 );
 
-=head2 inbox
-
-L<Coro::Channel> instanced used to queue tasks from the C<queue> method.
-
-=cut
-
-has inbox => (
-    is      => 'lazy',
-    isa     => InstanceOf['Coro::Channel'],
-);
-
-sub _build_inbox {
-  my $self = shift;
-  return Coro::Channel->new($self->max_procs);
-}
-
-=head2 inbox_worker
-
-L<Coro> thread monitoring the C<inbox>.
-
-=cut
-
-has inbox_worker => (
-    is  => 'lazy',
-    isa => InstanceOf['Coro'],
-);
-
-sub _build_inbox_worker {
-    async {
-        my $self = shift;
-
-        while (1) {
-            my $task = $self->inbox->get or last;
-            my ($f, $args, $on_success, $on_error) = @$task;
-
-            async_pool {
-                my ($self, $on_success, $on_error) = @_;
-                my $result = eval { $self->process($f, $args) };
-
-                if ($@) {
-                    my $error = $@;
-                    if (ref $on_error && ref $on_error eq 'CODE') {
-                        $on_error->($error);
-                    }
-                }
-                else {
-                    if (ref $on_success && ref $on_success eq 'CODE') {
-                        $on_success->($result);
-                    }
-                }
-            } $self, $on_success, $on_error;
-        }
-    } @_;
-}
-
 =head2 is_running
 
 Boolean which signals to the instance that the C<shutdown> method has been
@@ -196,7 +142,7 @@ called.
 has is_running => (
     is      => 'rw',
     isa     => Bool,
-    default => 1,
+    default => sub { 1 },
 );
 
 =head1 METHODS
@@ -210,7 +156,9 @@ sub DEMOLISH {
 
 sub BUILD {
     my $self = shift;
-    $self->inbox_worker; # ensure the worker starts
+    for (1 .. $self->max_procs) {
+      $self->checkin_proc($self->start_proc);
+    }
 }
 
 sub start_proc {
@@ -231,6 +179,7 @@ sub checkin_proc {
 
     if (!$self->is_running || ($self->max_reqs && $proc->messages_sent >= $self->max_reqs)) {
         $self->kill_proc($proc);
+        unshift @{$self->procs}, $self->start_proc;
     } else {
         unshift @{$self->procs}, $proc;
     }
@@ -275,9 +224,6 @@ sub shutdown {
 
     $self->is_running(0);
 
-    $self->inbox->shutdown;
-    $self->inbox_worker->join;
-
     my $count = $self->num_procs or return;
     for (1 .. $count) {
         my $proc = shift @{$self->procs};
@@ -306,10 +252,8 @@ C<max_reqs> can cause this method to yield while a new process is spawned.
 
 sub process {
     my ($self, $f, $args) = @_;
-
     my $guard = $self->procs_lock->guard;
-
-    my $proc = $self->checkout_proc;
+    my $proc  = $self->checkout_proc;
     scope_guard { $self->checkin_proc($proc) };
 
     my $msgid = $proc->send($f, $args);
@@ -356,39 +300,38 @@ sub defer {
     return sub { $cv->recv };
 }
 
-=head2 queue($f, $args, $callback)
+=head2 pipeline
 
-Queues the execution of C<$f->(@$args)> and returns immediately. If
-C<$callback> is specified and is a code ref, it will be called with the result
-of the executed code once the task is processed. Note that the callback is not
-provided with any identifying information about the task being executed. That
-is the responsibility of the caller. It is also the caller's responsibility to
-coordinate any code that depends on the result, for example using a condition
-variable.
+Returns a L<Coro::ProcessPool::Pipeline> object which can be used to pipe
+requests through to the process pool. Results then come out the other end of
+the pipe.  It is up to the calling code to perform task account (for example,
+by passing an id in as one of the arguments to the task class).
 
-    my $cv = AnyEvent->condvar;
+    my $pipe = $pool->pipeline;
 
-    sub make_callback {
-        my $n = shift;
-        return sub {
-            my $result = shift;
-            print "2 * $n = $result\n";
-            $cv->send;
+    my $producer = async {
+        foreach my $args (@tasks) {
+            $pipe->queue('Some::Class', $args);
         }
+
+        $pipe->shutdown;
+    };
+
+    while (my $result = $pipe->next) {
+        ...
     }
 
-    $pool->queue($doubler_function, [21], make_callback(21));
-    $cv->recv;
+All arguments to C<pipeline()> are passed transparently to the constructor of
+L<Coro::ProcessPool::Pipeline>. There is no limit to the number of pipelines
+which may be created for a pool.
 
 =cut
 
-sub queue {
-    my ($self, $f, $args, $on_success, $on_error) = @_;
-    $self->inbox->put([$f, $args, $on_success, $on_error]);
+sub pipeline {
+    my $self = shift;
+    return Coro::ProcessPool::Pipeline->new(@_);
 }
 
-=back
-
 =head1 A NOTE ABOUT IMPORTS AND CLOSURES
 
 Code refs are serialized using L<Storable> to pass them to the worker
@@ -493,6 +436,8 @@ process pool on Windows:
 
 =item L<Win32::Pipe>
 
+=back
+
 =head1 AUTHOR
 
 Jeff Ober <jeffober@gmail.com>
